@@ -24,7 +24,8 @@ class QwilFuzzer():
     OPT_POST = "post"
     OPT_PROB = "prob"
     OPT_PROBFB = "probfb"
-    DEFAULT_OPTS = [OPT_PRE, OPT_POST, OPT_PROB, OPT_PROBFB]
+    OPT_MAB = "mab"
+    DEFAULT_OPTS = [OPT_PRE, OPT_POST, OPT_PROB, OPT_PROBFB, OPT_MAB]
     DEFAULT_MIN_NONTERMINALS = 0
     DEFAULT_MAX_NONTERMINALS = 1000
 
@@ -35,6 +36,10 @@ class QwilFuzzer():
     # Minimum probability for entries being decreased
     PROB_MIN = 0.001
 
+    # Epsilon used for epsilon-greedy MAB solver
+    MAB_EPSILON = 0.1
+
+
     def __init__(self,
                  grammar,
                  start_symbol=DEFAULT_START_SYMBOL,
@@ -47,6 +52,7 @@ class QwilFuzzer():
         self.min_nonterminals = min_nonterminals
         self.max_nonterminals = max_nonterminals
         self.unguided_fuzzing = unguided_fuzzing
+        self.top_score = (0, self.grammar.copy())
 
         self.supported_opts = set()
         if supported_opts:
@@ -181,6 +187,35 @@ class QwilFuzzer():
         for c in children:
             self.update_grammar_probabilities(c)
 
+    def record_score(self, score):
+        if score > self.top_score[0]:
+            log.info(f"Updating score, old: {self.top_score[0]}, new: {score}")
+            self.top_score = (score, self.grammar.copy())
+
+    def update_mab_switches(self):
+        # Updating the grammar has been disabled for this instance, return
+        if self.unguided_fuzzing:
+            return
+
+        if random.random() <= QwilFuzzer.MAB_EPSILON: # Explore
+            log.info("Fuzzer going into explore mode")
+            self.randomize_mab_switch_states()
+        else: # Exploit
+            log.info("Fuzzer going into exploit mode")
+            self.grammar = self.top_score[1].copy()
+
+    def get_mab_switches(self):
+        mabs = {}
+
+        for nonterminal in self.grammar:
+            for expansion in self.grammar[nonterminal]:
+                mab = self.opt_mab(expansion) # include fixed and dynamic
+                if mab != None:
+                    mabs.update(
+                        {f"{nonterminal}.{expansion_string(expansion)}": mab})
+
+        return mabs
+
 ###############################################################################
 ############################################################### Private methods
 ###############################################################################
@@ -212,7 +247,12 @@ class QwilFuzzer():
 
         log.debug("Expanding %s randomly", all_terminals(node))
 
-        expansions = self.grammar[symbol]
+        expansions = self.filter_mab_switches(self.grammar[symbol])
+
+        # Expand into empty string if no available expansions
+        if len(expansions) == 0:
+            return (symbol, [("", [])])
+
         children_alternatives = \
             [self.expansion_to_children(expansion) for expansion in expansions]
 
@@ -233,6 +273,9 @@ class QwilFuzzer():
     def opt_prob(self, expansion):
         return expansion_opt(expansion, QwilFuzzer.OPT_PROB)
 
+    def opt_mab(self, expansion):
+        return expansion_opt(expansion, QwilFuzzer.OPT_MAB)
+
     def opt_prob_all(self, expansion):
         prob = expansion_opt(expansion, QwilFuzzer.OPT_PROB)
 
@@ -457,7 +500,7 @@ class QwilFuzzer():
             raise DerivationTreeError("Unexpandable node, children: " +
                                       children)
 
-        expansions = self.grammar[symbol]
+        expansions = self.filter_mab_switches(self.grammar[symbol])
 
         children_alternatives_with_cost = [
             (self.expansion_to_children(expansion),
@@ -544,3 +587,12 @@ class QwilFuzzer():
                                ": couldn't set default probbilities")
 
         return all_probs
+
+    def filter_mab_switches(self, expansions):
+        return [e for e in expansions if self.opt_mab(e) != 0]
+
+    def randomize_mab_switch_states(self):
+        for _,expansions in self.grammar.items():
+            for e in expansions:
+                if self.opt_mab(e) != None:
+                    e[EXP_OPTS_INDEX][QwilFuzzer.OPT_MAB] = random.randint(0,1)
